@@ -1,4 +1,4 @@
-# Copyright (c) 1998 by Jonathan Swartz. All rights reserved.
+# Copyright (c) 1998-99 by Jonathan Swartz. All rights reserved.
 # This program is free software; you can redistribute it and/or modify it
 # under the same terms as Perl itself.
 
@@ -10,6 +10,8 @@ require Exporter;
 @EXPORT_OK = ();
 
 use mod_perl;
+use strict;
+use vars (qw($AUTOLOAD));
 
 sub new {
     my $class = shift;
@@ -49,17 +51,19 @@ sub access_table {
 	return $self->{$field};
 	# die "FakeApache is unable to simulate the Apache::Table class for the `$field' method called in scalar context.  You'll need to temporarily remove or comment out that method call to use debug files.\n";
     } elsif (@_==3) {
-	return $self->{$key};
+	return $self->{$field}->{$key};
     } else {
-	return ($self->{$key} = $value);
+	return ($self->{$field}->{$key} = $value);
     }
 }
 
 sub args {
+    my $self = shift;
     return (wantarray) ? @{$self->{'args@'}} : $self->{'args$'};
 }
 
 sub get_basic_auth_pw {
+    my $self = shift;
     return @{$self->{get_basic_auth_pw}};
 }
 
@@ -87,12 +91,13 @@ sub warn {
 sub send_http_header {
     my $self = shift;
     $self->content_type('text/plain') if !$self->content_type;
-    print "Server: ".$ENV{'SERVER_SOFTWARE'}."\n" if $ENV{'SERVER_SOFTWARE'};
-    print "Content-type: ".$self->content_type."\n";
+    my $outm = $self->{headers_out_method} || sub { print $_[0] };
+    $outm->("Server: ".$ENV{'SERVER_SOFTWARE'}."\n") if $ENV{'SERVER_SOFTWARE'};
+    $outm->("Content-type: ".$self->content_type."\n");
     my $href = $self->headers_out;
     if (ref($href) eq 'HASH') {
 	while (my ($key,$value) = each(%{$href})) {
-	    print "$key: $value\n";
+	    $outm->("$key: $value\n");
 	}
     }
     print "\n";
@@ -130,6 +135,7 @@ sub AUTOLOAD {
 }
 
 package HTML::Mason::FakeApache::Server;
+use vars (qw($AUTOLOAD));
 sub new {
     my $class = shift;
     my $self = {
@@ -165,6 +171,7 @@ sub AUTOLOAD {
 }
 
 package HTML::Mason::FakeApache::Connection;
+use vars (qw($AUTOLOAD));
 sub new {
     my $class = shift;
     my $self = {
